package org.jamesii.experimentgeneration.controller.generation.controller;


import freemarker.cache.ClassTemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateExceptionHandler;
import org.apache.commons.lang3.StringUtils;
import org.jamesii.experimentgeneration.controller.generation.controller.backend.BackendController;
import org.jamesii.experimentgeneration.controller.generation.controller.backend.PythonBackendController;
import org.jamesii.experimentgeneration.controller.generation.controller.backend.SesslBackendController;
import org.jamesii.experimentgeneration.controller.generation.controller.backend.YamlBackendController;
import org.jamesii.experimentgeneration.model.BackendType;
import org.jamesii.experimentgeneration.model.ExperimentModel;
import org.jamesii.experimentgeneration.model.ExperimentType;
import org.jamesii.experimentgeneration.view.View;
import org.json.JSONObject;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;

/**
 * Created by Marcus on 17.12.2018.
 */
public class MainExperimentController implements Observer {
    private final String OS = System.getProperty("os.name").toLowerCase();
    private ParsingController parsingController;
    private PythonBackendController pythonBackendController;
    private SesslBackendController sesslBackendController;
    private Configuration configuration;
    private ExperimentModel experimentModel;
    private BackendController backendController;
    private YamlBackendController yamlBackendController;
    private View view;

    public MainExperimentController(ExperimentModel experimentModel, View view) {
        this.experimentModel = experimentModel;
        parsingController = new ParsingController(this);
        pythonBackendController = new PythonBackendController(this);
        sesslBackendController = new SesslBackendController(this);
        yamlBackendController = new YamlBackendController(this);
        experimentModel.addObserver(this);
        this.view = view;


            this.configuration = configureTemplateEngine();


        view.getStage().show();


        view.getGenerateExperimentButton().setOnAction(e -> {
                    experimentModel.generateExperiment(view);
                    if (parsingController.schemaCheckExperimentParameter(experimentModel.getSchemaFile(), experimentModel.getJsonInput())) {
                        forwardToBackend(experimentModel.getJsonInput());
                        Template template = selectTemplate();
                        renderTemplate(experimentModel, template);

                        try {

                            executeExperiment(experimentModel);
                        } catch (IOException ex) {
                            ex.printStackTrace();
                        }


                    }
                }
        );


    }

    private Template selectTemplate() {
        //TODO adjust to SESSL-ML ...
        Template template = null;
        ExperimentType experimentType = experimentModel.getExperimentType();
        BackendType backendType = experimentModel.getBackendType();
        if (backendType == BackendType.SESSL && experimentType == ExperimentType.STOCHASTIC_DISCRETE_EVENT) {
            try {
                template = configuration.getTemplate("Stochastic-Experiments/stochastic_master.ftl");
                return template;
            } catch (IOException e) {
                e.printStackTrace();
            }

        }else if (backendType == BackendType.YAML && experimentType == ExperimentType.FINITE_ELEMENT){
            try {
                template = configuration.getTemplate("Finite-Element/electromagnetic_master.ftl");
                return template;
            } catch (IOException e){
                e.printStackTrace();
            }
        }
        return template;
    }


    public View getView() {
        return view;
    }

    public ParsingController getParsingController() {
        return parsingController;
    }

    public ExperimentModel getExperimentModel() {
        return experimentModel;
    }

    public Configuration getConfiguration() {
        return configuration;
    }

    private void renderTemplate(ExperimentModel experimentModel, Template template) {
        switch (experimentModel.getBackendType()) {
            case PYTHON: {
                pythonBackendController.generateExperiment(experimentModel, template);
                break;
            }
            case SESSL: {
                sesslBackendController.generateExperiment(experimentModel, template);
                break;
            }
            case YAML: {
                yamlBackendController.generateExperiment(experimentModel, template);
                break;
            }
        }
    }

    private void forwardToBackend(JSONObject experimentInputs) {
        HashMap experimentParameter = experimentModel.getExperimentParameter();
        experimentParameter.clear();
        switch (parsingController.identifyBackend(experimentInputs)) {
            case PYTHON:
//                if (pythonBackendController.check(experimentInputs)) {
//                    HashMap experimentParameter = pythonBackendController.getExperimentParameter();
//                    return pythonBackendController.adjustModelPath(experimentParameter);
//
//                }
            case SESSL: {
                sesslBackendController.extractExperimentParameter(experimentParameter, experimentInputs);
                experimentModel.setBackendType(BackendType.SESSL);
//                //Perform Backend checks
                if (sesslBackendController.check(experimentInputs)) {
                    experimentModel.extractExperimentSpecification(experimentParameter);
                }
                break;
            }
            case YAML: {
                yamlBackendController.extractExperimentParameter(experimentParameter, experimentInputs);
                experimentModel.setBackendType(BackendType.YAML);
//                //Perform Backend checks
                if (yamlBackendController.check(experimentInputs)) {
                    experimentModel.extractExperimentSpecification(experimentParameter);
                }
                break;
            }
        }

    }

    private Configuration configureTemplateEngine() {
        Configuration cfg = new Configuration();

        ClassTemplateLoader loader = new ClassTemplateLoader(
                    this.getClass(), "/Templates");
        cfg.setTemplateLoader(loader);
        cfg.setDefaultEncoding("UTF-8");
        cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);

           // cfg.setDirectoryForTemplateLoading(new File(".\\src\\main\\resources\\Templates"));
            //cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
            return cfg;

    }

    private void executeExperimentBatch() throws IOException {
        ProcessBuilder processBuilder = new ProcessBuilder("cmd", "/C", "run.bat");
        File dir = new File("sessl-quickstart");
        if (dir.exists()){
            System.out.println("Directory Found");
        }
        processBuilder.directory(dir);
        System.out.println("Running Experiment");

        processBuilder.start();

    }

    private void executeExperimentShell() {
        try {
            ProcessBuilder pb = new ProcessBuilder("sh", "-c", "./run.sh");
            pb.redirectErrorStream();
            pb.directory(new File("sessl-quickstart"));
            pb.start();
            System.out.println("test");
        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    private void adjustPomFile(File experimentFile) throws IOException {
        File pomFile = new File("sessl-quickstart/pom.xml");
        String pomString = new String(Files.readAllBytes(Paths.get("sessl-quickstart/pom.xml")));
        String[] split = pomString.split(System.getProperty("line.separator"));
        for (int i = 0; i < split.length; i++) {
            System.out.println(split[i]);
            if (split[i].contains("scriptFile")) {
                split[i] = "<scriptFile>" + experimentFile.getName() + "</scriptFile>";
            }
        }
        List<String> wordList = Arrays.asList(split);
        String joined = StringUtils.join(wordList, System.getProperty("line.separator"));
        BufferedWriter writer = new BufferedWriter(new FileWriter(pomFile));
        writer.write(joined);
        writer.close();


    }

    public void executeExperiment(ExperimentModel experimentModel) throws IOException {
        String experimentName = experimentModel.getExperimentSpecification().get("experimentName").toString();
        switch (experimentModel.getBackendType()) {
            case SESSL: {
                writeExperimentFile(experimentModel);
                File experimentFile = createSesslExperimentFile(experimentName);
                adjustPomFile(experimentFile);
                if (isWindows()) {
                    executeExperimentBatch();
                } else if (isUnix()) {
                    executeExperimentShell();
                }
                break;
            }
            case PYTHON: {
                break;
            }
            case YAML: {

            }

        }
    }

    private File createSesslExperimentFile(String experimentName) {
        return new File("sessl-quickstart/" + experimentName + ".scala");
    }

    private boolean isWindows() {

        return (OS.contains("win"));

    }

    private boolean isUnix() {

        return (OS.contains("nix") || OS.contains("nux") || OS.indexOf("aix") > 0);
    }

    private void writeExperimentFile(ExperimentModel experimentModel) {
        try {
            String experimentDescription = experimentModel.getExperimentDescription();
            String experimentName = experimentModel.getExperimentSpecification().get("experimentName").toString();

            PrintWriter printWriter = new PrintWriter(new File("sessl-quickstart/" + experimentName + ".scala"));
            printWriter.write(experimentDescription);
            printWriter.flush();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }

    }

    @Override
    public void update(Observable o, Object arg) {

    }
}
