package org.jamesii.experimentgeneration.view;


import com.scalified.tree.TreeNode;
import com.scalified.tree.multinode.ArrayMultiTreeNode;
import javafx.event.EventHandler;
import javafx.geometry.HPos;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.*;
import javafx.stage.Stage;
import javafx.util.Pair;
import org.jamesii.experimentgeneration.model.ExperimentModel;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.*;

/**
 * Created by Marcus on 19.06.2019.
 */


public class View implements Observer {

    private ExperimentModel experimentModel;
    private Stage stage;
    private BorderPane outerBorderPane;
    private BorderPane innerBorderPane;
    private TabPane experimentTabs;
    private HBox buttonArea;
    private EditorView editorView;
    private ExperimentTab experimentTab;
    private Label loggerLabel;
    private VBox loggerView;
    private TextArea logger;
    private Button generateExperimentButton;
    private ExperimentMenu experimentMenu;
    public View(ExperimentModel experimentModel, Stage stage) {

        this.experimentModel = experimentModel;
        experimentModel.addObserver(this);
        this.experimentMenu = new ExperimentMenu(this);
        this.stage = stage;
        experimentTab = new ExperimentTab(experimentModel, this);
        editorView = new EditorView(experimentModel);
        outerBorderPane = new BorderPane();
        innerBorderPane = new BorderPane();
        experimentTabs = new TabPane();
        buttonArea = new HBox();

        this.loggerLabel = new Label("Schema Validation Error Log");
        this.loggerView = new VBox();
        this.logger = new TextArea();

        generateExperimentButton = new Button("Generate Experiment");

        setup();
    }

    public ExperimentModel getExperimentModel() {
        return experimentModel;
    }

    public ExperimentMenu getExperimentMenu() {
        return experimentMenu;
    }

    public TextArea getLogger() {
        return logger;
    }

    public Button getGenerateExperimentButton() {
        return generateExperimentButton;
    }

    public EditorView getEditorView() {
        return editorView;
    }

    public HBox getButtonArea() {
        return buttonArea;
    }

    public BorderPane getOuterBorderPane() {
        return outerBorderPane;
    }

    public BorderPane getInnerBorderPane() {
        return innerBorderPane;
    }

    public TabPane getExperimentTabs() {
        return experimentTabs;
    }

    public void setup() {
        buttonArea.setStyle("-fx-border-color: black;");
        outerBorderPane.setBottom(innerBorderPane);
        outerBorderPane.setStyle("-fx-border-color: black;");

        logger.setPrefWidth(400);
        loggerView.getChildren().add(loggerLabel);

        loggerView.getChildren().add(logger);
        outerBorderPane.setTop(experimentMenu);


        VBox.setVgrow(logger, Priority.ALWAYS);
        outerBorderPane.setRight(loggerView);



        innerBorderPane.setCenter(experimentTabs);
        outerBorderPane.setCenter(editorView);

        Scene scene = new Scene(outerBorderPane, 1024, 720);
        stage.setResizable(true);
        stage.setScene(scene);

        experimentTabs.getTabs().add(experimentTab);

        createButtonArea();
    }

    private void createButtonArea() {
        buttonArea.getChildren().add(generateExperimentButton);
        buttonArea.setPadding(new Insets(20));
        buttonArea.setAlignment(Pos.CENTER_RIGHT);
        buttonArea.setSpacing(20);
        innerBorderPane.setBottom(buttonArea);
    }

    /**
     * Generates Input User Interface based on selected Experiment Schema
     *
     * @param schema
     */
    public void generateInputElements(File schema) {
        experimentTab.getExperimentDesign().setDisable(false);
        HashMap<String, Tab> tabHash = experimentModel.getExperimentTabs();
        HashMap<String, Object> generatedGuiElements = experimentModel.getGeneratedGuiElements();
        LinkedHashMap<String, Object> map = experimentModel.getJsonAsLinkedHashMap(schema);
        TreeNode<String> experimentRoot = experimentModel.getExperimentRoot();

        for (String key : map.keySet()) {
            //For every Property
            if (key.equals("properties")) {
                if (map.get(key) instanceof LinkedHashMap) {
                    LinkedHashMap<String, Object> properties = (LinkedHashMap<String, Object>) map.get(key);
                    //For all Properties , e.g Inputs
                    for (String propertyKey : properties.keySet()) {

                        LinkedHashMap<String, Object> item = (LinkedHashMap<String, Object>) properties.get(propertyKey);
                        //for every item!

                        for (String itemKeys : item.keySet()) {
                            String tabText = item.get("tab").toString();
                            if (!tabText.isEmpty()) {
                                createNewTab(tabHash, tabText);
                            }
                            if (itemKeys.equals("type")) {
                                //Generate GUI depending on type
                                switch (item.get(itemKeys).toString()) {
                                    case "string":
                                        //Label + Textfield
                                    case "integer":
                                    case "number": {
                                        TreeNode<String> property = new ArrayMultiTreeNode<String>(propertyKey);
                                        experimentRoot.add(property);
                                        generateBasicElements(tabHash, generatedGuiElements, propertyKey, tabText, item, map);
                                        break;
                                    }
                                    //Label with multiple Textfields
                                    case "array": {
                                        TreeNode<String> property = new ArrayMultiTreeNode<String>(propertyKey);
                                        experimentRoot.add(property);
                                        generateDynamicInputFields(tabHash, propertyKey, tabText, item, generatedGuiElements);
                                        break;
                                    }
                                    //New Tab recursively
                                    case "object": {
                                        TreeNode<String> property = new ArrayMultiTreeNode<String>(propertyKey);
                                        experimentRoot.add(property);
                                        handleObjectCase(tabHash, generatedGuiElements, item, tabText, map, property);
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        experimentModel.updateDisabledElementsHash(map);
        experimentModel.initialDisable();
    }


    private void handleObjectCase(HashMap<String, Tab> tabHash, HashMap<String, Object> generatedGuiElements, LinkedHashMap<String, Object> item, String tabText, LinkedHashMap<String, Object> map, TreeNode<String> property) {
        LinkedHashMap<String, Object> itemObject = (LinkedHashMap<String, Object>) item.get("properties");
        for (String objectKey : itemObject.keySet()) {
            LinkedHashMap<String, Object> object = (LinkedHashMap<String, Object>) itemObject.get(objectKey);
            for (String values : object.keySet()) {
                if (values.equals("type")) {
                    switch (object.get(values).toString()) {
                        case "string":
                        case "number":
                        case "integer": {
                            TreeNode<String> propertyChild = new ArrayMultiTreeNode<String>(objectKey);
                            property.add(propertyChild);
                            generateBasicElements(tabHash, generatedGuiElements, objectKey, tabText, object, map);
                            break;
                        }
                        case "array": {
                            TreeNode<String> propertyChild = new ArrayMultiTreeNode<String>(objectKey);
                            property.add(propertyChild);
                            generateDynamicInputFields(tabHash, objectKey, tabText, object, generatedGuiElements);
                            break;
                        }
                        case "object": {
                            TreeNode<String> propertyChild = new ArrayMultiTreeNode<String>(objectKey);
                            property.add(propertyChild);
                            handleObjectCase(tabHash, generatedGuiElements, object, tabText, map, propertyChild);
                            break;
                        }
                    }
                }
            }
        }
    }


    private void generateDynamicInputFields(HashMap<String, Tab> tabHash, String propertyKey, String tabText, LinkedHashMap<String, Object> item, HashMap<String, Object> generatedGuiElement) {
        HashMap<String, Pair<Integer, Integer>> zipMarker = experimentModel.getZipInstanceMarker();
        HashMap<String, Pair<Button, Button>> zipButtonMarker = experimentModel.getZipButtonMarker();
        LinkedHashMap<String, LinkedHashMap<String, ArrayList<Node>>> zipHash = experimentModel.getZipHash();
        HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker = experimentModel.getGridPositionMarker();
        LinkedHashMap<String, ArrayList> zipInstances = experimentModel.getZipInstances();

        String zip = "";
        if (item.containsKey("zipName")) {
            zip = item.get("zipName").toString();
        }
        String finalZip = zip;

        Tab tab = tabHash.get(tabText);
        ScrollPane scrollPane = (ScrollPane) tab.getContent();
        GridPane layout = (GridPane) scrollPane.getContent();

        Label label = setPropertyLabel(propertyKey, item);
        TextField textField = generateNewTextField(propertyKey, item);
        Button addButton = createAddButton(propertyKey, finalZip, tab);
        Button removeButton = createRemoveButton(zip, tab);


        Pair<Integer, Integer> indexPair = gridPositionMarker.get(tab);
        Integer rowIndex = indexPair.getValue();
        Integer columnIndex = indexPair.getKey();

        //If gridpane is empty
        if (rowIndex == 0 && columnIndex == 0) {
            addElementToEmptyPane(propertyKey, item, generatedGuiElement, zipMarker, zipButtonMarker, zipHash, gridPositionMarker, zipInstances, label, textField, addButton, removeButton, tab, layout, rowIndex);

        }
        //If Gridpane is not Empty
        else {
            addElementToNonEmptyPane(propertyKey, item, generatedGuiElement, zipMarker, zipButtonMarker, zipHash, gridPositionMarker, zipInstances, label, textField, addButton, removeButton, tab, layout, rowIndex, columnIndex, zip);

        }
    }

    @NotNull
    private Button createAddButton(String propertyKey, String finalZip, Tab tab) {
        Button addButton = new Button("Add");
        addButton.setOnAction(event -> {
            experimentModel.pushGridElementsDown(tab, addButton);
            experimentModel.addNewRow(tab, addButton, finalZip, propertyKey);
        });
        return addButton;
    }

    @NotNull
    private Button createRemoveButton(String zip, Tab tab) {
        Button removeButton = new Button("Remove");
        removeButton.setId(zip);
        return removeButton;
    }

    private void addElementToNonEmptyPane(String propertyKey, LinkedHashMap<String, Object> item, HashMap<String, Object> generatedGuiElement, HashMap<String, Pair<Integer, Integer>> zipMarker, HashMap<String, Pair<Button, Button>> zipButtonMarker, LinkedHashMap<String, LinkedHashMap<String, ArrayList<Node>>> zipHash, HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker, LinkedHashMap<String, ArrayList> zipInstances, Label label, TextField textField, Button addButton, Button removeButton, Tab tab, GridPane layout, Integer rowIndex, Integer columnIndex, String zip) {
        //if Element is part of a Zip
        if (item.containsKey("zipName")) {
            String zipName = item.get("zipName").toString();
            //if one zip part already exists and another is added
            updateZipStructure(propertyKey, zipInstances, zipName);

            if (zipMarker.containsKey(zipName)) {

                Pair<Integer, Integer> zipMark = zipMarker.get(zipName);
                Integer column = zipMark.getKey();
                Integer row = zipMark.getValue();

                Pair<Button, Button> buttonButtonPair = zipButtonMarker.get(zipName);
                Button key = buttonButtonPair.getKey();
                Button value = buttonButtonPair.getValue();

                layout.getChildren().remove(key);
                layout.getChildren().remove(value);

                layout.add(label, column, row);
                layout.add(textField, column + 1, row);
                layout.add(addButton, column + 2, row);
                layout.add(removeButton, column + 3, row);

                updateZipStructure(propertyKey, zipHash, textField, zipName);
                zipMarker.put(zipName, new Pair<>(column + 2, row));
                zipButtonMarker.put(zipName, new Pair<>(addButton, removeButton));
            } else
            //if it's the first zip part
            {
                //Update Zip structure
                createNewZipInstance(propertyKey, zipHash, textField, zip);

                layout.add(label, columnIndex, rowIndex + 1);
                layout.add(textField, columnIndex + 1, rowIndex + 1);
                layout.add(addButton, columnIndex + 2, rowIndex + 1);
                layout.add(removeButton, columnIndex + 3, rowIndex + 1);


                zipButtonMarker.put(zipName, new Pair<>(addButton, removeButton));
                gridPositionMarker.put(tab, new Pair<>(0, rowIndex + 1));
                zipMarker.put(zipName, new Pair<>(columnIndex + 2, rowIndex + 1));
            }
        }
        //If element is not part of a zip and layout is not empty
        else {
            ArrayList<Node> list = new ArrayList<>();
            list.add(textField);
            generatedGuiElement.put(propertyKey, list);

            addGuiElementsToLayout(gridPositionMarker, label, textField, addButton, removeButton, tab, layout, rowIndex, columnIndex, rowIndex + 1, columnIndex + 1, columnIndex + 2, columnIndex + 3);
        }
    }

    private void addElementToEmptyPane(String propertyKey, LinkedHashMap<String, Object> item, HashMap<String, Object> generatedGuiElement, HashMap<String, Pair<Integer, Integer>> zipMarker, HashMap<String, Pair<Button, Button>> zipButtonMarker, LinkedHashMap<String, LinkedHashMap<String, ArrayList<Node>>> zipHash, HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker, LinkedHashMap<String, ArrayList> zipInstances, Label label, TextField textField, Button addButton, Button removeButton, Tab tab, GridPane layout, Integer rowIndex) {
        addGuiElementsToLayout(gridPositionMarker, label, textField, addButton, removeButton, tab, layout, rowIndex, 0, 0, 1, 2, 3);

        //N-Tuple Zip
        if (item.containsKey("zipName")) {
            String zipName = item.get("zipName").toString();


            updateZipStructure(propertyKey, zipInstances, zipName);
            createNewZipInstance(propertyKey, zipHash, textField, zipName);

            zipMarker.put(zipName, new Pair<>(2, 0));
            zipButtonMarker.put(zipName, new Pair<>(addButton, removeButton));
        } else {
            ArrayList<Node> nodes = new ArrayList<>();
            nodes.add(textField);
            generatedGuiElement.put(propertyKey, nodes);
        }
    }

    private void addGuiElementsToLayout(HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker, Label label, TextField textField, Button addButton, Button removeButton, Tab tab, GridPane layout, Integer rowIndex, int i, int i2, int i3, int i4, int i5) {
        layout.add(label, i, i2);
        layout.add(textField, i3, i2);
        GridPane.setHalignment(addButton, HPos.LEFT);
        layout.add(addButton, i4, i2);
        GridPane.setHalignment(removeButton, HPos.LEFT);
        layout.add(removeButton, i5, i2);
        gridPositionMarker.put(tab, new Pair<>(0, rowIndex + 1));
    }

    private void updateZipStructure(String propertyKey, LinkedHashMap<String, ArrayList> zipInstances, String zipName) {
        if (!zipInstances.containsKey(zipName)) {
            ArrayList zipList = new ArrayList();
            zipList.add(propertyKey);
            zipInstances.put(zipName, zipList);
        } else {
            ArrayList arrayList = zipInstances.get(zipName);
            arrayList.add(propertyKey);
        }
    }

    private TextField generateNewTextField(String propertyKey, LinkedHashMap<String, Object> item) {
        TextField textField = new TextField();
        textField.setId(propertyKey);
        textField.setPromptText(getLabelValue(propertyKey, item));
        return textField;
    }


    private void updateZipStructure(String propertyKey, LinkedHashMap<String, LinkedHashMap<String, ArrayList<Node>>> zipHash, TextField textField, String zipName) {
        LinkedHashMap<String, ArrayList<Node>> zipInstance = zipHash.get(zipName);
        ArrayList<Node> additionalElementList = new ArrayList<>();
        additionalElementList.add(textField);
        zipInstance.put(propertyKey, additionalElementList);
    }

    private void createNewZipInstance(String propertyKey, LinkedHashMap<String, LinkedHashMap<String, ArrayList<Node>>> zipHash, TextField textField, String zip) {
        LinkedHashMap<String, ArrayList<Node>> individualZip = new LinkedHashMap<>();
        LinkedHashMap<String, Object> generatedGuiElements = experimentModel.getGeneratedGuiElements();
        ArrayList<Node> propertyElements = new ArrayList<>();
        propertyElements.add(textField);
        individualZip.put(propertyKey, propertyElements);
        zipHash.put(zip, individualZip);
        generatedGuiElements.put(zip, individualZip);

    }

    private Label setPropertyLabel(String propertyKey, LinkedHashMap<String, Object> item) {
        String labelText = getLabelValue(propertyKey, item);
        return new Label(labelText);
    }


    private void generateBasicElements(HashMap<String, Tab> tabHash, HashMap<String, Object> generatedGuiElements, String propertyKey, String tabText, LinkedHashMap<String, Object> item, LinkedHashMap<String, Object> map) {
        String labelText = getLabelValue(propertyKey, item);
        Label label = new Label(labelText);


        Tab tab = tabHash.get(tabText);
        ScrollPane scrollPane = (ScrollPane) tab.getContent();
        GridPane layout = (GridPane) scrollPane.getContent();

        HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker = experimentModel.getGridPositionMarker();
        Pair<Integer, Integer> indexPair = gridPositionMarker.get(tab);
        Integer rowIndex = indexPair.getValue();
        Integer columnIndex = indexPair.getKey();

        if (item.containsKey("enum")) {
            generateEnumerationElements(generatedGuiElements, propertyKey, item, label, tab, layout, gridPositionMarker, indexPair, rowIndex, columnIndex, map);
        } else if (!item.containsKey("enum")) {
            generateInputTextField(generatedGuiElements, propertyKey, label, tab, layout, gridPositionMarker, indexPair, rowIndex, columnIndex, item);
        }
    }

    private String getLabelValue(String propertyKey, LinkedHashMap<String, Object> item) {
        String labelText;
        if (item.containsKey("label")) {
            labelText = item.get("label").toString();
        } else {
            labelText = propertyKey;
        }
        return labelText;
    }

    private void generateInputTextField(HashMap<String, Object> generatedGuiElements, String propertyKey, Label label, Tab tab, GridPane layout, HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker, Pair<Integer, Integer> indexPair, Integer rowIndex, Integer columnIndex, LinkedHashMap<String, Object> item) {
        TextField textField = generateNewTextField(propertyKey, item);
        generatedGuiElements.put(propertyKey, textField);
        //If Textfield is the first element in the tab
        if (indexPair.getKey() == 0 && rowIndex == 0) {
            layout.add(label, 0, 0);
            layout.add(textField, 1, 0);
            Pair<Integer, Integer> updatedPair = new Pair<>(0, rowIndex + 1);
            gridPositionMarker.put(tab, updatedPair);
        }
        //if tab is not empty start with new row
        else if (indexPair.getKey() == 0 && rowIndex != 0) {
            layout.add(label, columnIndex, rowIndex + 1);
            layout.add(textField, columnIndex + 1, rowIndex + 1);
            Pair<Integer, Integer> updatedPair = new Pair<>(0, rowIndex + 1);
            gridPositionMarker.put(tab, updatedPair);
        }
    }

    private void generateEnumerationElements(HashMap<String, Object> generatedGuiElements, String propertyKey, LinkedHashMap<String, Object> item, Label label, Tab tab, GridPane layout, HashMap<Tab, Pair<Integer, Integer>> gridPositionMarker, Pair<Integer, Integer> indexPair, Integer rowIndex, Integer columnIndex, LinkedHashMap<String, Object> map) {
        ArrayList<Object> enumArray = (ArrayList<Object>) item.get("enum");
        ComboBox<String> comboBox = new ComboBox<>();
        for (Object enumerationValues : enumArray) {
            comboBox.getItems().add(enumerationValues.toString());
        }
        generatedGuiElements.put(propertyKey, comboBox);
        comboBox.setId(propertyKey);
        comboBox.getSelectionModel().select(0);
        if (indexPair.getKey() == 0 && rowIndex == 0) {
            layout.add(label, 0, 0);
            layout.add(comboBox, 1, 0);
            Pair<Integer, Integer> updatedPair = new Pair<>(0, 1);
            gridPositionMarker.put(tab, updatedPair);
        }
        //if tab is not empty start with new row
        else if (indexPair.getKey() == 0 && rowIndex != 0) {
            layout.add(label, columnIndex, rowIndex + 1);
            layout.add(comboBox, columnIndex + 1, rowIndex + 1);
            Pair<Integer, Integer> updatedPair = new Pair<>(0, rowIndex + 1);
            gridPositionMarker.put(tab, updatedPair);
        }


        comboBox.setOnAction(e -> {
            System.out.print(comboBox.getSelectionModel().getSelectedItem());
            experimentModel.updateEdibility(generatedGuiElements, comboBox);
        });
    }


    private void createNewTab(HashMap<String, Tab> tabHashMap, String text) {

        Tab tab = new Tab(text);
        tab.setClosable(false);
        ScrollPane scrollPane = new ScrollPane();
        scrollPane.setPrefHeight(300);
        experimentModel.getGridPositionMarker().put(tab, new Pair<>(0, 0));
        GridPane gridPane = new GridPane();
        gridPane.setHgap(10);
        gridPane.setVgap(10);
        gridPane.setPadding(new Insets(10, 10, 10, 10));
        gridPane.setVisible(true);
        scrollPane.setContent(gridPane);
        tab.setContent(scrollPane);

        if (!tabHashMap.containsKey(tab.getText())) {
            tabHashMap.put(tab.getText(), tab);
            experimentTabs.getTabs().add(tab);
        }

    }

    public Stage getStage() {
        return stage;
    }

    @Override
    public void update(Observable o, Object arg) {

    }

    public void clearView() {
        ArrayList<Tab> tabs = new ArrayList<>();
        for (Tab tab : experimentTabs.getTabs()) {
            if (tab != experimentTab) {
                tabs.add(tab);
            }
        }
        for (Tab todeleteTab : tabs) {
            experimentTabs.getTabs().remove(todeleteTab);
        }
    }




}
